/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2008, 2009. All Rights Reserved.
 *  
 * U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 ******************************************************************************/
package com.ibm.rational.test.lm.spi.virutualization.mock;

import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import com.ibm.ccl.soa.deploy.core.HostingLink;
import com.ibm.ccl.soa.deploy.core.Unit;
import com.ibm.ccl.soa.deploy.net.IpInterfaceUnit;
import com.ibm.ccl.soa.deploy.operation.RunState;
import com.ibm.ccl.soa.deploy.os.OperatingSystemUnit;
import com.ibm.ccl.soa.deploy.server.Server;
import com.ibm.ccl.soa.deploy.server.ServerUnit;
import com.ibm.ccl.soa.deploy.virtualization.VirtualImageCollectionUnit;
import com.ibm.ccl.soa.deploy.virtualization.VirtualImageUnit;
import com.ibm.rational.test.lm.spi.common.SPIAuthenticationException;
import com.ibm.rational.test.lm.spi.common.SPIClientConnectionException;
import com.ibm.rational.test.lm.spi.common.SPIException;
import com.ibm.rational.test.lm.spi.common.SPIResourceNotFoundException;
import com.ibm.rational.test.lm.spi.util.topology.TopologyCreator;
import com.ibm.rational.test.lm.spi.util.topology.TopologyConstant.RemoteConnectionTypes;

public class VirtualizationService {

	// singleton
	public static final VirtualizationService INSTANCE = new VirtualizationService();

	private static HashMap<String, ServerUnit> virtualMachines = new HashMap<String, ServerUnit>();
	private static HashMap<String, VirtualImageUnit> virtualImages = new HashMap<String, VirtualImageUnit>();
	private static HashMap<String, VirtualImageCollectionUnit> virtualImageCollections = new HashMap<String, VirtualImageCollectionUnit>();
	private static HashMap<String, Unit> virtualMachineCollections = new HashMap<String, Unit>();
	
	// Init sample virtual machines
	// Each machine unit need to link with deployed image unit
	public static void initVM() {
		for (int i = 0; i < 4; i++) {
			Object[] keys = virtualImages.keySet().toArray();
			// Get the first virtual image for sample
			VirtualImageUnit sampleVI = virtualImages.get(keys[0]);
			String vmId = "VM_Sample" + i;
			String vmName = "VirtualMachine_" + vmId;
	        OperatingSystemUnit osUnit = TopologyCreator.createOSUnit("AIX", "AIX", vmName);
	        String ip = "127.0.0." + i;
	        IpInterfaceUnit ipInterfaceUnit = TopologyCreator.createIpInterfaceUnit(ip, false);
			ServerUnit virtualMachine = TopologyCreator
					.createVirtualMachineUnitForVMId(vmName, vmId, "Sample VM Description", 
					        RunState.RUNNING_LITERAL, osUnit, ipInterfaceUnit,
					        RemoteConnectionTypes.HTTP, "http://www.mylivelink.com");
			// link the machine unit with the image unit
			TopologyCreator.createHostLink(virtualMachine, sampleVI);
			virtualMachines.put(vmId, virtualMachine);
		}
	}
	
	private static String[] viTypes = {
	    "Microsoft Hyper-V", //$NON-NLS-N$	
        "Microsoft Virtual Server", //$NON-NLS-N$
        "VMware ESX", //$NON-NLS-N$
        "VMware Workstation", //$NON-NLS-N$
        "XenSource"  //$NON-NLS-N$
        };
	
	// Init sample virtual images
	public static void initVI() {
		for (int i = 0; i < 4; i++) {
			String viId = "VI_Sample" + i;
			String viName = "VirtualImage_" + viId; 
			OperatingSystemUnit osUnit = 
			    TopologyCreator.createOSUnit("AIX", "AIX", viId);
			VirtualImageUnit virtualImage = TopologyCreator
					.createVirtualImageUnit(viName, viId, "Sample Virtual Image", 
					        viTypes[i], "VMWare ESX", "1.0", 
					        new Timestamp(System.currentTimeMillis()), 
					        new Timestamp(System.currentTimeMillis()), osUnit,
					        RemoteConnectionTypes.FTP, "ftp://myimagestorage.com"); //$NON-NLS-N$
			virtualImages.put(viId, virtualImage);
		}
	}
	
	// Init sample virtual images collections
	public static void initVIC() {	
	    String viCollectionName = "VIC_Sample";
	    String viCollectionId = "1";
		VirtualImageCollectionUnit viCollectionUnit = TopologyCreator
				.createVirtualImageCollectionUnit(viCollectionName, viCollectionId, "Sample Image Collection Description");
		int size = virtualImages.keySet().size();
		Object[] keys = virtualImages.keySet().toArray();
		for (int j = 0; j < size; j++) {
			VirtualImageUnit viUnit = virtualImages.get(keys[j]);
			TopologyCreator.createHostLink(viCollectionUnit, viUnit);
		}
		virtualImageCollections.put(viCollectionId, viCollectionUnit);
		
	}

	// Init sample virtual machine collections
	// Each machine collection unit need to link with the deployed image collection unit
	// Each machine collection unit need to link with its machine units
	// Each machine unit need to link with its deployed image unit
	public static void initVMC() {
		String vmCollectionId = "VMC_Sample";
		String vmCollectionName = "VirtualMachineCollection_" + vmCollectionId;
		Object[] viColkeys = virtualImageCollections.keySet().toArray();
		// Get the first virtual image collection
		VirtualImageCollectionUnit viCollectionUnit = virtualImageCollections.get(viColkeys[0]);
		VirtualImageUnit[] viUnits = getVirtualImageFromCollection(viCollectionUnit);
		// Create virtual machine collection unit
		Unit vmCollectionUnit = TopologyCreator
				.createVirtualMachineCollectionUnit(
				        vmCollectionName, vmCollectionId, "Sample VM Collection Description", RunState.CREATED_LITERAL);
		//Create Sample virtual machines from virtual image
		int ipCount=0;
		for (VirtualImageUnit viUnit : viUnits) {
			String machineId = "VirtualMachine_" + viUnit.getName();
			String vmName =  machineId;
	        OperatingSystemUnit osUnit = TopologyCreator.createOSUnit("AIX", "AIX", vmName);
	        String ip = "127.0.1." + ipCount;
	        ipCount++;
	        IpInterfaceUnit ipInterfaceUnit = TopologyCreator.createIpInterfaceUnit(ip, false);
			// The machine's states is running
			ServerUnit virtualMachineUnit = TopologyCreator
					.createVirtualMachineUnitForVMId(
					        vmName, machineId, "Sample VM from Collection Description", 
					        	RunState.CREATED_LITERAL, osUnit, ipInterfaceUnit,
					        	RemoteConnectionTypes.HTTP, "http://www.mylivelink.com");
			// Associate the image unit with machine unit
			TopologyCreator.createHostLink(virtualMachineUnit, viUnit);
			// Associate the machine unit with machine collection
			TopologyCreator.createHostLink(vmCollectionUnit, virtualMachineUnit);
			
			virtualMachines.put(machineId, virtualMachineUnit);
		}	
		// Associate the image collection unit with machine collection
		TopologyCreator.createHostLink(vmCollectionUnit, viCollectionUnit);
		virtualMachineCollections.put(vmCollectionId, vmCollectionUnit);

	}

	/**
	 * Get all virtual images' detail information from mock data
	 * 
	 * @return
	 */
	public String getVirtualImages() {
		List<VirtualImageUnit> list = new ArrayList<VirtualImageUnit>();
		Iterator<String> iter = virtualImages.keySet().iterator();
		while (iter.hasNext()) {
			VirtualImageUnit virtualImageUnit = virtualImages.get(iter.next());
			list.add(virtualImageUnit);
		}
		String virtualImageTopology = TopologyCreator
				.createVirtualImagesTopology(list);
		return virtualImageTopology;
	}

	/**
	 * Get all virtual machines' detail information from mock data
	 * 
	 * @return
	 */
	public String getVirtualMachines() {
		List<ServerUnit> virtualMachineUnits = new ArrayList<ServerUnit>();
		Iterator<String> iter = virtualMachines.keySet().iterator();
		while (iter.hasNext()) {
			ServerUnit virtualMachineUnit = virtualMachines.get(iter.next());
			virtualMachineUnits.add(virtualMachineUnit);
		}
		String virtualMachineTopology = TopologyCreator
				.createVirtualMachinesTopology(virtualMachineUnits);
		return virtualMachineTopology;
	}

	/**
	 * Get virtual machine detail information by its id
	 * 
	 * @param resourceId
	 * @return
	 */
	public String getVirtualMachine(String resourceId) {
		List<ServerUnit> virtualMachineUnits = new ArrayList<ServerUnit>(1);
		ServerUnit virtualMachine = virtualMachines.get(resourceId);
		if (virtualMachine != null) {
			virtualMachineUnits.add(virtualMachine);
		}
		String virtualMachineTopology = TopologyCreator
				.createVirtualMachinesTopology(virtualMachineUnits);
		return virtualMachineTopology;

	}

	/**
	 * Delete (Undeploy) a virtual machine by its id
	 * 
	 * @param resourceId
	 * @param providerHostName
	 * @param providerPort
	 * @return
	 */
	public boolean deleteVirtualMachine(String resourceId) {
		boolean deleteSuccess = false;
		if (virtualMachines.containsKey(resourceId)) {
			virtualMachines.remove(resourceId);
			// Remove the virtual machine from virtual machine collection that
			// it is a part of
			removeVMFromContainedVmCols(virtualMachineCollections
					.values().toArray(
							new Unit[virtualMachineCollections.size()]),
					resourceId);
			deleteSuccess = true;
		}
		return deleteSuccess;
	}
	
	/**
	 * Deploy virtual image and return virtual machine id
	 * 
	 * @param XMLBody
	 * @return
	 */
	public String deployImage(String resourceId) throws SPIException {
		/*
		 * In the real implementation of deploying a virtual image, the steps
		 * may include:
		 * 
		 * 1. Establish connection to real virtualization provider by its API.
		 * It is possible to encounter connection error or authentification
		 * error.
		 * 
		 * 2. Find the virtual image in the repository of the provider. If it
		 * isn't found, the error should be return back.
		 * 
		 * 3. Execute the deploying operation, any exception should be caught,
		 * such as service exception, permission exception, etc.
		 * 
		 * 4. Return the id of created virtual machine in TDM.
		 * 
		 */
		if (virtualImages.containsKey(resourceId)) {
			// The size of virtualImages is 4 in the mock
			List<String> viIds = new ArrayList<String>(virtualImages.size());
			viIds.addAll(virtualImages.keySet());
			int index = viIds.indexOf(resourceId);
			// The reason of error, it need to be wrapped readable for end user.
			// TODO How to localize error message in provider adaptor
			// implementation?
			String errMsg = null;
			switch (index) {
			case 0:
				try {
					// Deploy the virtual image and return virtual machine TDM
					
					// Mock the machineId to change each time
					String machineId = resourceId + "_Deploy_in_"
							+ System.currentTimeMillis();
					String vmName = "VirtualMachine_" + machineId;
					List<ServerUnit> virtualMachineUnits = new ArrayList<ServerUnit>(
							1);
					OperatingSystemUnit osUnit = TopologyCreator.createOSUnit(
							"AIX", "AIX", vmName);
					String ip = "127.0.2." + System.currentTimeMillis() % 254;
					IpInterfaceUnit ipInterfaceUnit = TopologyCreator
							.createIpInterfaceUnit(ip, false);
					// First the machine's states is Created
					ServerUnit virtualMachine = TopologyCreator
							.createVirtualMachineUnitForVMId(vmName, machineId,
									"Sample VM Description",
									RunState.CREATED_LITERAL, osUnit,
									ipInterfaceUnit,
									RemoteConnectionTypes.HTTP, "http://www.mylivelink.com");
					// Get the first virtual image from resourceId
					VirtualImageUnit virtualImage = virtualImages
							.get(resourceId);
					// create link to link image unit with machine unit
					TopologyCreator
							.createHostLink(virtualMachine, virtualImage);
					virtualMachineUnits.add(virtualMachine);
					String virtualMachineTopology = TopologyCreator
							.createVirtualMachinesTopology(virtualMachineUnits);
					// Change the machine's status to Running
					TopologyCreator.changeVMRunState(virtualMachine,
							RunState.RUNNING_LITERAL);
					virtualMachines.put(machineId, virtualMachine);
					return virtualMachineTopology;
				} catch (Exception e) {
					// Catch internal error of the provide and wrap the reason
					// into SPIException
					errMsg = "Encountered service exception on the remote provider";
					throw new SPIException(errMsg, e);
				}
			case 1:
				// Throw connection exception
				errMsg = "Timeout while trying to communicate with the remote provider server";
				throw new SPIClientConnectionException(errMsg);
			case 2:
				// Throw authentification exception
				errMsg = "Failed to authenticate with the remote provider server";
				throw new SPIAuthenticationException(errMsg);
			case 3:
			default:
				// Throw a SPIException, it could be used for us to test
				// the error display on UI when running with the mock.
				errMsg = "The virtual image can't be deployed for permission denied"; 
				throw new SPIException(errMsg);
			}
		} else {
			// The virtual image hasn't existed on the provider yet, throw a
			// SPIResourceNotFoundException.
			String errMsg = "The virtual image hasn't existed on the provider yet";
			throw new SPIResourceNotFoundException(errMsg);
		}
	}

	/**
	 * Deploy virtual images collection and return the virtual machines 
	 * collection external id TDM xml string. 
	 * 
	 * @param resourceId
	 * @return
	 */
	public String deployVirtualImageCollection(String resourceId) {
		// Create the deployed virtual machine collection id
		String vmCollExId = resourceId + System.currentTimeMillis();
		String vmCollName = "VirtualMachineCollection_" + vmCollExId;
		List<Unit> vmCollectionUnits = new ArrayList<Unit>(1);
		// Get the virtual image collection by image collection id
		VirtualImageCollectionUnit viCollectionUnit = virtualImageCollections
				.get(resourceId);
		// Create the virtual machine collectio unit and set the RunState to create for mock
		Unit vmCollIdUnit = TopologyCreator.
		    createVirtualMachineCollectionUnit(vmCollName, vmCollExId,
		    		"Sample Virtual Machine Collection Description", RunState.CREATED_LITERAL);
		
		// Get all the virtual images's information
		VirtualImageUnit[] viUnits = getVirtualImageFromCollection(viCollectionUnit);
		
		//since we'll be increasing ip, need room to grow to 254
		int ipCount = (int)System.currentTimeMillis() % 250;  
		for (VirtualImageUnit viUnit : viUnits) {
			String machineId = viUnit.getName() + System.currentTimeMillis();
			String vmName = "VirtualMachine_" + machineId;
	        OperatingSystemUnit osUnit = TopologyCreator.createOSUnit("AIX", "AIX", vmName);
	        String ip = "127.0.3." + ipCount;
	        ipCount++;
	        IpInterfaceUnit ipInterfaceUnit = TopologyCreator.createIpInterfaceUnit(ip, false);
			// Create the virutal machine for each image
			ServerUnit virtualMachineUnit = TopologyCreator
					.createVirtualMachineUnitForVMId(
					        vmName, machineId, "Sample Virtual Machine Unit Description", 
					        	RunState.CREATED_LITERAL, osUnit, ipInterfaceUnit,
					        	RemoteConnectionTypes.HTTP, "http://www.mylivelink.com");
			// Associate the image unit with machine unit
			TopologyCreator.createHostLink(virtualMachineUnit, viUnit);
			// Associate the machine unit with machine collection
			TopologyCreator.createHostLink(vmCollIdUnit, virtualMachineUnit);
			// Change the machine's status to Running
			TopologyCreator.changeVMRunState(virtualMachineUnit,
					RunState.RUNNING_LITERAL);
			// put the machine into mock data
			virtualMachines.put(machineId, virtualMachineUnit);
		}	
		// Associate the image collection unit with machine collection
		TopologyCreator.createHostLink(vmCollIdUnit, viCollectionUnit);
		
		vmCollectionUnits.add(vmCollIdUnit);
		// Create virtual machine collection by virtual image collection
		String top = TopologyCreator
				.createVirtualMachineCollectionTopology(vmCollectionUnits);
		
		// change state to running to make the deploying finish.
		TopologyCreator.changeVMRunState(vmCollIdUnit,
				RunState.RUNNING_LITERAL);
		
		//put the vm collection into mock data
		virtualMachineCollections.put(vmCollExId, vmCollIdUnit);

		return top;
	}

	/**
	 * Get all virtual ImageCollections' detail information from mock data
	 * 
	 * @return
	 */
	public String getVirtualImageCollections() {
		List<VirtualImageCollectionUnit> virtualImageCollectionUnits = new ArrayList<VirtualImageCollectionUnit>();
		Iterator<String> iter = virtualImageCollections.keySet().iterator();
		while (iter.hasNext()) {
			VirtualImageCollectionUnit virtualImageCollectionUnit = virtualImageCollections
					.get(iter.next());
			virtualImageCollectionUnits.add(virtualImageCollectionUnit);
		}
		String virtualImageCollectionTopology = TopologyCreator
				.createVirtualImageCollectionTopology(virtualImageCollectionUnits);
		return virtualImageCollectionTopology;
	}

	/**
	 * Get virtual Image Collection' detail information by its Ids
	 * 
	 * @param resourceId
	 * @return
	 */
	public String getVirtualImageCollection(String resourceId) {
		List<VirtualImageCollectionUnit> virtualImageCollectionUnits = new ArrayList<VirtualImageCollectionUnit>(
				1);
		VirtualImageCollectionUnit virtualCollection = virtualImageCollections
				.get(resourceId);
		virtualImageCollectionUnits.add(virtualCollection);
		String virtualImageCollectionTopology = TopologyCreator
				.createVirtualImageCollectionTopology(virtualImageCollectionUnits);
		return virtualImageCollectionTopology;
	}

	/**
	 * Get all virtual Image Collections' Ids
	 * 
	 * @return
	 */
	public String getVirtualImageCollectionIds() {
		return getVirtualImageCollections();
	}

	/**
	 * Get all virtual ImageCollections' detail information
	 * 
	 * @return
	 */
	public String getVirtualMachineCollections() {
		List<Unit> virtualMachineCollectionUnits = new ArrayList<Unit>();
		Iterator<String> iter = virtualMachineCollections.keySet().iterator();
		while (iter.hasNext()) {
			Unit virtualMachineCollectionUnit = virtualMachineCollections
					.get(iter.next());
			virtualMachineCollectionUnits.add(virtualMachineCollectionUnit);
		}
		String virtualMachineCollectionTopology = TopologyCreator
				.createVirtualMachineCollectionTopology(virtualMachineCollectionUnits);
		return virtualMachineCollectionTopology;
	}
	
	/**
	 * Get virtual Machine Collection' detail information by its Id
	 * 
	 * @param resourceId
	 * @return
	 */
	public String getVirtualMachineCollection(String resourceId) {
		List<Unit> virtualMachineCollectionUnits = new ArrayList<Unit>();
		String virtualMachineCollectionTopology = null;
		Unit virtualMachineCollection = virtualMachineCollections
				.get(resourceId);
		if (virtualMachineCollection != null) {
		virtualMachineCollectionUnits.add(virtualMachineCollection);
		virtualMachineCollectionTopology = TopologyCreator
				.createVirtualMachineCollectionTopology(virtualMachineCollectionUnits);
		}
		return virtualMachineCollectionTopology;
	}

	/**
	 * Get all virtual Machine Collections' Ids
	 * 
	 * @return
	 */
	public String getVirtualMachineCollectionIds() {
		return getVirtualMachineCollections();
	}
	
	/**
	 * Delete (Undeploy) virtual machine collection and contained virtual
	 * machines by the given collection id
	 * 
	 * @param resourceId
	 * @return
	 */
	@SuppressWarnings("unchecked")
	public boolean deleteVirtualMachineCollection(String resourceId) {
		boolean deleteSuccess = false;
		if (virtualMachineCollections.containsKey(resourceId)) {
			// Delete all of virtual machines in the collection
			Unit vmColUnit = virtualMachineCollections.get(resourceId);
			String[] vmExternalIds = getVMExternalIdsFromVMCol(vmColUnit);
			for (String vmExternalId : vmExternalIds) {
				deleteVirtualMachine(vmExternalId);
			}
			// Deleted the collection
			virtualMachineCollections.remove(resourceId);
			deleteSuccess = true;
		}
		return deleteSuccess;
	}
	
	   /**
     * Remove the virtual machine unit and link from virtual machine collections
     * that it is a part of by the given external id of virtual machine
     * 
     * @param vmColUnits
     * @param vmExternalId
     */
    @SuppressWarnings("unchecked")
    private static void removeVMFromContainedVmCols(Unit[] vmColUnits,
            String vmExternalId) {
        for (Unit vmColUnit : vmColUnits) {
            List<HostingLink> links = vmColUnit.getHostingLinks();
            int removedLinkIndex = -1;
            for (int index = 0; index < links.size(); index++) {
                Unit unit = links.get(index).getHosted();
                if (unit instanceof ServerUnit) {
                    ServerUnit serverUnit = (ServerUnit) unit;
                    Server server = (Server) serverUnit.getCapabilities()
                            .get(0);
                    if (vmExternalId.equals(server.getSystemId())) {
                        removedLinkIndex = index;
                        break;
                    }
                }
            }
            if (removedLinkIndex >= 0) {
                vmColUnit.getHostingLinks().remove(removedLinkIndex);
            }
        }
    }
	
    /**
     * Parse all of external id of virtual machines from the given virtual
     * machine collection unit
     * 
     * @param vmColUnit
     * @return Array of external ids of virtual machines in the collection
     */
    @SuppressWarnings("unchecked")
    private static String[] getVMExternalIdsFromVMCol(Unit vmColUnit) {
        List<HostingLink> links = vmColUnit.getHostingLinks();
        ArrayList<String> vmExternalIds = new ArrayList<String>();
        for (HostingLink link : links) {
            Unit unit = link.getHosted();
            if (unit instanceof ServerUnit) {
                ServerUnit serverUnit = (ServerUnit) unit;
                Server server = (Server) serverUnit.getCapabilities().get(0);
                vmExternalIds.add(server.getSystemId());
            }
        }
        return vmExternalIds.toArray(new String[vmExternalIds.size()]);
    }    
    
    /**
     * Get the Virtual Image Units linking with Virtual Image Collection Unit
     * 
     * @param Virtual
     *            Image Collection Unit
     * @return
     */
    private static VirtualImageUnit[] getVirtualImageFromCollection(
            VirtualImageCollectionUnit collectionUnit) {
        List<VirtualImageUnit> vImageUnits = new ArrayList<VirtualImageUnit>();
        collectionUnit.getHostingLinks();
        List<?> allLinks = collectionUnit.getHostingLinks();
        for (Object link : allLinks) {
            Unit targetUnit = ((HostingLink) link).getTarget();
            if (targetUnit instanceof VirtualImageUnit) {
                vImageUnits.add((VirtualImageUnit) targetUnit);
            }
        }
        return vImageUnits.toArray(new VirtualImageUnit[vImageUnits.size()]);
    }
}
